/*
 *  npw-viewer.c - Target plugin loader and viewer
 *
 *  nspluginwrapper (C) 2005-2006 Gwenole Beauchesne
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _GNU_SOURCE 1 /* RTLD_NEXT */
#include "sysdeps.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <pthread.h>
#include <sys/poll.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/extensions/XShm.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "rpc.h"
#include "npw-rpc.h"
#include "utils.h"

#define XP_UNIX 1
#define MOZ_X11 1
#include <npapi.h>
#include <npupp.h>

#define DEBUG 1
#include "debug.h"


// [UNIMPLEMENTED] Define to use XPCOM emulation
#define USE_XPCOM 0

// [UNIMPLEMENTED] Define to use XEMBED emulation
#define USE_XEMBED 0

// Define to dispatch all Xt events within the main GTK thread, including signals & timers
#define DISPATCH_ALL_XT_EVENTS 0

// XXX unimplemented functions
#define UNIMPLEMENTED() npw_printf("WARNING: Unimplemented function %s at line %d\n", __func__, __LINE__)


// RPC global connections
static rpc_connection_t *g_rpc_connection;

// Instance state information about the plugin
typedef struct _PluginInstance {
  NPP instance;
  uint32_t npn_instance;
  Window old_window;
#if USE_XEMBED
  GtkWidget *window;
#else
  Window window;
#endif
  Widget top_widget;
  Widget form;
  int width, height;
} PluginInstance;


/* ====================================================================== */
/* === X Toolkit glue                                                 === */
/* ====================================================================== */

static Display *x_display;
static XtAppContext x_app_context;

typedef struct _XtTMRec {
    XtTranslations  translations;       /* private to Translation Manager    */
    XtBoundActions  proc_table;         /* procedure bindings for actions    */
    struct _XtStateRec *current_state;  /* Translation Manager state ptr     */
    unsigned long   lastEventTime;
} XtTMRec, *XtTM;   

typedef struct _CorePart {
    Widget          self;               /* pointer to widget itself          */
    WidgetClass     widget_class;       /* pointer to Widget's ClassRec      */
    Widget          parent;             /* parent widget                     */
    XrmName         xrm_name;           /* widget resource name quarkified   */
    Boolean         being_destroyed;    /* marked for destroy                */
    XtCallbackList  destroy_callbacks;  /* who to call when widget destroyed */
    XtPointer       constraints;        /* constraint record                 */
    Position        x, y;               /* window position                   */
    Dimension       width, height;      /* window dimensions                 */
    Dimension       border_width;       /* window border width               */
    Boolean         managed;            /* is widget geometry managed?       */
    Boolean         sensitive;          /* is widget sensitive to user events*/
    Boolean         ancestor_sensitive; /* are all ancestors sensitive?      */
    XtEventTable    event_table;        /* private to event dispatcher       */
    XtTMRec         tm;                 /* translation management            */
    XtTranslations  accelerators;       /* accelerator translations          */
    Pixel           border_pixel;       /* window border pixel               */
    Pixmap          border_pixmap;      /* window border pixmap or NULL      */
    WidgetList      popup_list;         /* list of popups                    */
    Cardinal        num_popups;         /* how many popups                   */
    String          name;               /* widget resource name              */
    Screen          *screen;            /* window's screen                   */
    Colormap        colormap;           /* colormap                          */
    Window          window;             /* window ID                         */
    Cardinal        depth;              /* number of planes in window        */
    Pixel           background_pixel;   /* window background pixel           */
    Pixmap          background_pixmap;  /* window background pixmap or NULL  */
    Boolean         visible;            /* is window mapped and not occluded?*/
    Boolean         mapped_when_managed;/* map window if it's managed?       */
} CorePart;

typedef struct _WidgetRec {
    CorePart    core;
} WidgetRec, CoreRec;


/* ====================================================================== */
/* === XPCOM glue                                                     === */
/* ====================================================================== */

#if defined(__GNUC__) && (__GNUC__ > 2)
#define NS_LIKELY(x)    (__builtin_expect((x), 1))
#define NS_UNLIKELY(x)  (__builtin_expect((x), 0))
#else
#define NS_LIKELY(x)    (x)
#define NS_UNLIKELY(x)  (x)
#endif

#define NS_FAILED(_nsresult) (NS_UNLIKELY((_nsresult) & 0x80000000))
#define NS_SUCCEEDED(_nsresult) (NS_LIKELY(!((_nsresult) & 0x80000000)))

typedef uint32 nsresult;
typedef struct nsIServiceManager nsIServiceManager;
extern nsresult NS_GetServiceManager(nsIServiceManager **result);


/* ====================================================================== */
/* === Browser side plug-in API                                       === */
/* ====================================================================== */

static char *g_user_agent = NULL;

static inline uint32_t instance_id(NPP instance)
{
  return (uintptr_t)instance->ndata;
}

// Netscape exported functions
static NPNetscapeFuncs mozilla_funcs;

// Closes and deletes a stream
static NPError
g_NPN_DestroyStream(NPP instance, NPStream *stream, NPError reason)
{
  D(bug("NPN_DestroyStream instance=%p\n", instance));

  UNIMPLEMENTED();

  return NPERR_GENERIC_ERROR;
}

// Forces a repaint message for a windowless plug-in
static void
g_NPN_ForceRedraw(NPP instance)
{
  D(bug("NPN_ForceRedraw instance=%p\n", instance));

  UNIMPLEMENTED();
}

// Asks the browser to create a stream for the specified URL
static NPError
invoke_NPN_GetURL(PluginInstance *plugin, const char *url, const char *target)
{
  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPN_GET_URL,
								RPC_TYPE_UINT32, plugin->npn_instance,
								RPC_TYPE_STRING, url,
								RPC_TYPE_STRING, target,
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPN_GetURL\n");
	return NPERR_GENERIC_ERROR;
  }

  int32_t ret;
  error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_INT32, &ret, RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NPN_GetURL\n");
	return NPERR_GENERIC_ERROR;
  }

  return ret;
}

static NPError
g_NPN_GetURL(NPP instance, const char *url, const char *target)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  PluginInstance *plugin = id_lookup(instance_id(instance));
  if (plugin == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  D(bug("NPN_GetURL instance=%p\n", instance));
  NPError ret = invoke_NPN_GetURL(plugin, url, target);
  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

// Requests creation of a new stream with the contents of the specified URL
static NPError
invoke_NPN_GetURLNotify(PluginInstance *plugin, const char *url, const char *target, void *notifyData)
{
  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPN_GET_URL_NOTIFY,
								RPC_TYPE_UINT32, plugin->npn_instance,
								RPC_TYPE_STRING, url,
								RPC_TYPE_STRING, target,
								RPC_TYPE_UINT32, notifyData,
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPN_GetURLNotify\n");
	return NPERR_GENERIC_ERROR;
  }

  int32_t ret;
  error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_INT32, &ret, RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NPN_GetURLNotify\n");
	return NPERR_GENERIC_ERROR;
  }

  return ret;
}

static NPError
g_NPN_GetURLNotify(NPP instance, const char *url, const char *target, void *notifyData)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  PluginInstance *plugin = id_lookup(instance_id(instance));
  if (plugin == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  D(bug("NPN_GetURLNotify instance=%p\n", instance));
  NPError ret = invoke_NPN_GetURLNotify(plugin, url, target, notifyData);
  D(bug(" return urlnotify: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

// Allows the plug-in to query the browser for information
static NPError
g_NPN_GetValue(NPP instance, NPNVariable variable, void *value)
{
  D(bug("NPN_GetValue instance=%p, variable=%d [%08x]\n", instance, variable & 0xffff, variable));

  switch (variable) {
  case NPNVxDisplay:
	*(void **)value = x_display;
	break;
  case NPNVxtAppContext:
	*(void **)value = XtDisplayToApplicationContext(x_display);
	break;
#if USE_XEMBED
	// XXX call into the browser
  case  NPNVSupportsXEmbedBool:
	*(PRBool *)value = PR_TRUE;
	break;
#endif
  case NPNVToolkit:
	*(NPNToolkitType *)value = NPNVGtk2;
	break;
#if USE_XPCOM
  case NPNVserviceManager: {
	nsIServiceManager *sm;
	int ret = NS_GetServiceManager(&sm);
	if (NS_FAILED(ret))
	  return NPERR_GENERIC_ERROR;
	*(nsIServiceManager **)value = sm;
	break;
  }
#endif
  default:
	npw_printf("WARNING: unhandled variable %d for NPN_GetValue\n", variable);
	return NPERR_INVALID_PARAM;
  }

  return NPERR_NO_ERROR;
}

// Invalidates specified drawing area prior to repainting or refreshing a windowless plug-in
static void
g_NPN_InvalidateRect(NPP instance, NPRect *invalidRect)
{
  D(bug("NPN_InvalidateRect instance=%p\n", instance));

  UNIMPLEMENTED();
}

// Invalidates specified region prior to repainting or refreshing a windowless plug-in
static void
g_NPN_InvalidateRegion(NPP instance, NPRegion invalidRegion)
{
  D(bug("NPN_InvalidateRegion instance=%p\n", instance));

  UNIMPLEMENTED();
}

// Allocates memory from the browser's memory space
static void *
g_NPN_MemAlloc(uint32 size)
{
  D(bug("NPN_MemAlloc size=%d\n", size));

  return malloc(size);
}

// Requests that the browser free a specified amount of memory
static uint32
g_NPN_MemFlush(uint32 size)
{
  D(bug("NPN_MemFlush size=%d\n", size));

  return 0;
}

// Deallocates a block of allocated memory
static void
g_NPN_MemFree(void *ptr)
{
  D(bug("NPN_MemFree ptr=%p\n", ptr));

  free(ptr);
}

// Requests the creation of a new data stream produced by the plug-in and consumed by the browser
static NPError
g_NPN_NewStream(NPP instance, NPMIMEType type, const char *target, NPStream **stream)
{
  D(bug("NPN_NewStream instance=%p\n", instance));

  UNIMPLEMENTED();

  return NPERR_GENERIC_ERROR;
}

// Posts data to a URL
static NPError
g_NPN_PostURL(NPP instance, const char *url, const char *target, uint32 len, const char *buf, NPBool file)
{
  D(bug("NPN_PostURL instance=%p\n", instance));

  UNIMPLEMENTED();

  return NPERR_GENERIC_ERROR;
}

// Posts data to a URL, and receives notification of the result
static NPError
g_NPN_PostURLNotify(NPP instance, const char *url, const char *target, uint32 len, const char *buf, NPBool file, void *notifyData)
{
  D(bug("NPN_PostURLNotify instance=%p\n", instance));

  UNIMPLEMENTED();

  return NPERR_GENERIC_ERROR;
}

// Posts data to a URL, and receives notification of the result
static void
g_NPN_ReloadPlugins(NPBool reloadPages)
{
  D(bug("NPN_ReloadPlugins reloadPages=%d\n", reloadPages));

  UNIMPLEMENTED();
}

// Returns the Java execution environment
static JRIEnv *
g_NPN_GetJavaEnv(void)
{
  D(bug("NPN_GetJavaEnv\n"));

  return NULL;
}

// Returns the Java object associated with the plug-in instance
static jref
g_NPN_GetJavaPeer(NPP instance)
{
  D(bug("NPN_GetJavaPeer instance=%p\n", instance));

  return NULL;
}

// Requests a range of bytes for a seekable stream
static NPError
g_NPN_RequestRead(NPStream *stream, NPByteRange *rangeList)
{
  D(bug("NPN_RequestRead stream=%p\n", stream));

  UNIMPLEMENTED();

  return NPERR_GENERIC_ERROR;
}

// Sets various modes of plug-in operation
static NPError
g_NPN_SetValue(NPP instance, NPPVariable variable, void *value)
{
  D(bug("NPN_SetValue instance=%p\n", instance));

  UNIMPLEMENTED();

  return NPERR_GENERIC_ERROR;
}

// Displays a message on the status line of the browser window
static void
invoke_NPN_Status(PluginInstance *plugin, const char *message)
{
  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPN_STATUS,
								RPC_TYPE_UINT32, plugin->npn_instance,
								RPC_TYPE_STRING, message,
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPN_Status\n");
	return;
  }

  error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NPN_Status\n");
	return;
  }
}

static void
g_NPN_Status(NPP instance, const char *message)
{
  PluginInstance *plugin = id_lookup(instance_id(instance));
  if (plugin == NULL)
	return;

  D(bug("NPN_Status instance=%p\n", instance));
  invoke_NPN_Status(plugin, message);
  D(bug(" done\n"));
}

// Returns the browser's user agent field
static char *
invoke_NPN_UserAgent(void)
{
  int error = rpc_method_invoke(g_rpc_connection, RPC_METHOD_NPN_USER_AGENT, RPC_TYPE_INVALID);
  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPN_UserAgent\n");
	return NULL;
  }

  char *user_agent;
  error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_STRING, &user_agent, RPC_TYPE_INVALID);
  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NPN_UserAgent\n");
	return NULL;
  }

  return user_agent;
}

static const char *
g_NPN_UserAgent(NPP instance)
{
  D(bug("NPN_UserAgent instance=%p\n", instance));
  if (g_user_agent == NULL)
	g_user_agent = invoke_NPN_UserAgent();
  D(bug(" user_agent='%s'\n", g_user_agent));
  return g_user_agent;
}

// Pushes data into a stream produced by the plug-in and consumed by the browser
static int32
g_NPN_Write(NPP instance, NPStream *stream, int32 len, void *buf)
{
  D(bug("NPN_Write instance=%d\n", instance));

  UNIMPLEMENTED();

  return -1;
}


/* ====================================================================== */
/* === Plug-in side data                                              === */
/* ====================================================================== */

// Functions supplied by the plug-in
static NPPluginFuncs plugin_funcs;

// Allows the browser to query the plug-in supported formats
typedef char * (*NP_GetMIMEDescriptionUPP)(void);
static NP_GetMIMEDescriptionUPP g_NP_GetMIMEDescription = NULL;

// Allows the browser to query the plug-in for information
typedef NPError (*NP_GetValueUPP)(void *instance, NPPVariable variable, void *value);
static NP_GetValueUPP g_NP_GetValue = NULL;

// Provides global initialization for a plug-in
typedef NPError (*NP_InitializeUPP)(NPNetscapeFuncs *moz_funcs, NPPluginFuncs *plugin_funcs);
static NP_InitializeUPP g_NP_Initialize = NULL;

// Provides global deinitialization for a plug-in
typedef NPError (*NP_ShutdownUPP)(void);
static NP_ShutdownUPP g_NP_Shutdown = NULL;


/* ====================================================================== */
/* === RPC communication                                              === */
/* ====================================================================== */

// NP_GetMIMEDescription
static int handle_NP_GetMIMEDescription(rpc_connection_t *connection)
{
  D(bug("handle_NP_GetMIMEDescription\n"));

  char *str = g_NP_GetMIMEDescription();
  return rpc_method_send_reply(connection, RPC_TYPE_STRING, str, RPC_TYPE_INVALID);
}

// NP_GetValue
static int handle_NP_GetValue(rpc_connection_t *connection)
{
  D(bug("handle_NP_GetValue\n"));

  int32_t variable;
  int error = rpc_method_get_args(connection, RPC_TYPE_INT32, &variable, RPC_TYPE_INVALID);
  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NP_GetValue arguments\n");
	return error;
  }

  char *str = NULL;
  NPError ret = g_NP_GetValue ? g_NP_GetValue(NULL, variable, (void *)&str) : NPERR_GENERIC_ERROR;
  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_STRING, str, RPC_TYPE_INVALID);
}

// NP_Initialize
static int handle_NP_Initialize(rpc_connection_t *connection)
{
  D(bug("handle_NP_Initialize\n"));

  memset(&plugin_funcs, 0, sizeof(plugin_funcs));
  plugin_funcs.size = sizeof(plugin_funcs);

  memset(&mozilla_funcs, 0, sizeof(mozilla_funcs));
  mozilla_funcs.size = sizeof(mozilla_funcs);
  mozilla_funcs.version = (NP_VERSION_MAJOR << 8) + NP_VERSION_MINOR;
  mozilla_funcs.geturl = NewNPN_GetURLProc(g_NPN_GetURL);
  mozilla_funcs.posturl = NewNPN_PostURLProc(g_NPN_PostURL);
  mozilla_funcs.requestread = NewNPN_RequestReadProc(g_NPN_RequestRead);
  mozilla_funcs.newstream = NewNPN_NewStreamProc(g_NPN_NewStream);
  mozilla_funcs.write = NewNPN_WriteProc(g_NPN_Write);
  mozilla_funcs.destroystream = NewNPN_DestroyStreamProc(g_NPN_DestroyStream);
  mozilla_funcs.status = NewNPN_StatusProc(g_NPN_Status);
  mozilla_funcs.uagent = NewNPN_UserAgentProc(g_NPN_UserAgent);
  mozilla_funcs.memalloc = NewNPN_MemAllocProc(g_NPN_MemAlloc);
  mozilla_funcs.memfree = NewNPN_MemFreeProc(g_NPN_MemFree);
  mozilla_funcs.memflush = NewNPN_MemFlushProc(g_NPN_MemFlush);
  mozilla_funcs.reloadplugins = NewNPN_ReloadPluginsProc(g_NPN_ReloadPlugins);
  mozilla_funcs.getJavaEnv = NewNPN_GetJavaEnvProc(g_NPN_GetJavaEnv);
  mozilla_funcs.getJavaPeer = NewNPN_GetJavaPeerProc(g_NPN_GetJavaPeer);
  mozilla_funcs.geturlnotify = NewNPN_GetURLNotifyProc(g_NPN_GetURLNotify);
  mozilla_funcs.posturlnotify = NewNPN_PostURLNotifyProc(g_NPN_PostURLNotify);
  mozilla_funcs.getvalue = NewNPN_GetValueProc(g_NPN_GetValue);
  mozilla_funcs.setvalue = NewNPN_SetValueProc(g_NPN_SetValue);
  mozilla_funcs.invalidaterect = NewNPN_InvalidateRectProc(g_NPN_InvalidateRect);
  mozilla_funcs.invalidateregion = NewNPN_InvalidateRegionProc(g_NPN_InvalidateRegion);
  mozilla_funcs.forceredraw = NewNPN_ForceRedrawProc(g_NPN_ForceRedraw);

#if USE_XEMBED
  PRBool supports_XEmbed = PR_FALSE;

  NPError error = mozilla_funcs.getvalue(NULL, NPNVSupportsXEmbedBool, (void *)&supports_XEmbed);
  if (error != NPERR_NO_ERROR) {
	npw_printf("ERROR: your browser does not support the XEmbed extension\n");
	return NPERR_INCOMPATIBLE_VERSION_ERROR;
  }

  NPNToolkitType toolkit = 0;
  error = mozilla_funcs.getvalue(NULL, NPNVToolkit, (void *)&toolkit);
  if (error != NPERR_NO_ERROR || toolkit != NPNVGtk2) {
	npw_printf("ERROR: your browser does not support GTK+ 2 peer\n");
	return NPERR_INCOMPATIBLE_VERSION_ERROR;
  }
#endif

  NPError ret = g_NP_Initialize(&mozilla_funcs, &plugin_funcs);
  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_INVALID);
}

// NP_Shutdown
static int handle_NP_Shutdown(rpc_connection_t *connection)
{
  D(bug("handle_NP_Shutdown\n"));

  NPError ret = g_NP_Shutdown();
  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_INVALID);
}

// NP_Exit
static int handle_NP_Exit(rpc_connection_t *connection)
{
  D(bug("handle_NP_Exit\n"));

  gtk_main_quit();
  return RPC_ERROR_NO_ERROR;
}

// NPP_New
static int handle_NPP_New(rpc_connection_t *connection)
{
  D(bug("handle_NPP_New\n"));

  int i, error;
  NPMIMEType plugin_type;
  int32_t mode;
  int argn_count, argv_count;
  uint32_t npn_instance_id;
  char **argn, **argv;
  error = rpc_method_get_args(connection,
							  RPC_TYPE_UINT32, &npn_instance_id,
							  RPC_TYPE_STRING, &plugin_type,
							  RPC_TYPE_INT32, &mode,
							  RPC_TYPE_ARRAY, RPC_TYPE_STRING, &argn_count, &argn,
							  RPC_TYPE_ARRAY, RPC_TYPE_STRING, &argv_count, &argv,
							  RPC_TYPE_INVALID);
  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPP_New arguments\n");
	return error;
  }
  assert(argn_count == argv_count);

  PluginInstance *plugin = malloc(sizeof(PluginInstance));
  if (plugin == NULL)
	return RPC_ERROR_NO_MEMORY;
  memset(plugin, 0, sizeof(*plugin));

  plugin->instance = malloc(sizeof(NPP_t));
  if (plugin->instance == NULL)
	return RPC_ERROR_NO_MEMORY;
  memset(plugin->instance, 0, sizeof(*plugin->instance));
  plugin->npn_instance = npn_instance_id;
  int plugin_id = id_create(plugin);
  plugin->instance->ndata = (void *)(uintptr_t)plugin_id;

  NPError ret = plugin_funcs.newp(plugin_type, plugin->instance, mode, argn_count, argn, argv, NULL);
  free(plugin_type);
  if (argn) {
	for (i = 0; i < argn_count; i++)
	  free(argn[i]);
	free(argn);
  }
  if (argv) {
	for (i = 0; i < argv_count; i++)
	  free(argv[i]);
	free(argv);
  }

  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_UINT32, plugin_id, RPC_TYPE_INVALID);
}

// NPP_Destroy
static NPError g_NPP_Destroy(PluginInstance *plugin, NPSavedData **sdata)
{
  if (sdata)
	*sdata = NULL;

  if (plugin == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  if (plugin->window) {
	XSync(x_display, False);
#if USE_XEMBED
	XtUnregisterDrawable(x_display, GDK_WINDOW_XWINDOW(plugin->window->window));
#else
	XtUnregisterDrawable(x_display, plugin->window);
#endif
	XSync(x_display, False);
	plugin->top_widget->core.window = plugin->old_window;
	XtUnrealizeWidget(plugin->top_widget);
  }

  D(bug("NPP_Destroy instance=%p\n", plugin->instance));
  NPError ret = plugin_funcs.destroy(plugin->instance, sdata);
  free(plugin->instance);
  free(plugin);

  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

static int handle_NPP_Destroy(rpc_connection_t *connection)
{
  int error;
  uint32_t plugin_id;
  error = rpc_method_get_args(connection, RPC_TYPE_UINT32, &plugin_id, RPC_TYPE_INVALID);
  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPP_Destroy arguments\n");
	return error;
  }

  NPSavedData *save_area;
  NPError ret = g_NPP_Destroy(id_lookup(plugin_id), &save_area);
  id_unlink(plugin_id);

  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_NP_SAVED_DATA, save_area, RPC_TYPE_INVALID);
}

// NPP_SetWindow
static NPError
g_NPP_SetWindow(PluginInstance *plugin, NPWindow *window)
{
  if (plugin == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  if (plugin_funcs.setwindow == NULL)
	return NPERR_INVALID_FUNCTABLE_ERROR;

  D(bug("NPP_SetWindow instance=%p, window=%p\n", plugin->instance, window->window));
  NPError ret = plugin_funcs.setwindow(plugin->instance, window);
  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

static int handle_NPP_SetWindow(rpc_connection_t *connection)
{
  D(bug("handle_NPP_SetWindow\n"));

  int error;
  uint32_t plugin_id;
  NPWindow the_window, *window = &the_window;
  uint32_t window_id;
  uint32_t top, left, bottom, right;
  NPSetWindowCallbackStruct the_ws_info, *ws_info = &the_ws_info;
  int32_t visual_class;

  error = rpc_method_get_args(connection,
							  RPC_TYPE_UINT32, &plugin_id,
							  RPC_TYPE_UINT32, &window_id,
							  RPC_TYPE_INT32, &window->x,
							  RPC_TYPE_INT32, &window->y,
							  RPC_TYPE_UINT32, &window->width,
							  RPC_TYPE_UINT32, &window->height,
							  RPC_TYPE_INT32, &window->type,
							  RPC_TYPE_UINT32, &top,
							  RPC_TYPE_UINT32, &left,
							  RPC_TYPE_UINT32, &bottom,
							  RPC_TYPE_UINT32, &right,
							  RPC_TYPE_INT32, &ws_info->type,
							  RPC_TYPE_UINT32, &ws_info->depth,
							  RPC_TYPE_INT32, &visual_class,
							  RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPP_SetWindow arguments\n");
	return error;
  }

  window->window = (void *)(Window)window_id;
  window->clipRect.top = top;
  window->clipRect.left = left;
  window->clipRect.bottom = bottom;
  window->clipRect.right = right;
  window->ws_info = ws_info;

  ws_info->display = x_display;
  XVisualInfo visualInfo;
  int screen = DefaultScreen(ws_info->display);
  if (!XMatchVisualInfo(ws_info->display, screen, ws_info->depth, visual_class, &visualInfo)) {
	npw_printf("ERROR: could not reconstruct visual info for NPP_SetWindow\n");
	return RPC_ERROR_GENERIC;
  }
  ws_info->colormap = DefaultColormap(ws_info->display, screen);
  ws_info->visual = visualInfo.visual;

  PluginInstance *plugin = id_lookup(plugin_id);
  if (plugin->window) {
	if (plugin->width != window->width || plugin->height != window->height) {
	  plugin->width = window->width;
	  plugin->height = window->height;

	  XResizeWindow(x_display, XtWindow(plugin->form), plugin->width, plugin->height);
	  XResizeWindow(x_display, XtWindow(plugin->top_widget), plugin->width, plugin->height);

	  Arg args[2];
	  int n = 0;
	  XtSetArg(args[n], XtNwidth, plugin->width); n++;
	  XtSetArg(args[n], XtNheight, plugin->height); n++;
	  XtSetValues(plugin->form, args, n);
	  XtSetValues(plugin->top_widget, args, n);

	  XSync(x_display, False);
	}
  }
  else {
	plugin->width = window->width;
	plugin->height = window->height;
#if USE_XEMBED
	plugin->window = gtk_plug_new((Window)window->window);
	gtk_widget_set_usize(plugin->window, plugin->width, plugin->height);
	gtk_widget_show(plugin->window);
#else
	XSetWindowAttributes attr;
	attr.bit_gravity = NorthWestGravity;
	attr.colormap = ws_info->colormap;
	attr.event_mask =
	  ButtonMotionMask |
	  ButtonPressMask |
	  ButtonReleaseMask |
	  KeyPressMask |
	  KeyReleaseMask |
	  EnterWindowMask |
	  LeaveWindowMask |
	  PointerMotionMask |
	  StructureNotifyMask |
	  VisibilityChangeMask |
	  FocusChangeMask |
	  ExposureMask;

	unsigned long mask = CWBitGravity | CWEventMask;
	if (attr.colormap)
	  mask |= CWColormap;

	plugin->window = XCreateWindow(x_display, (Window)window->window,
								   0, 0, window->width, window->height,
								   0, ws_info->depth, InputOutput, ws_info->visual, mask, &attr);

	XSelectInput(x_display, plugin->window, ExposureMask);

	XMapWindow(x_display, plugin->window);
	XFlush(x_display);
#endif

	Widget top_widget = XtAppCreateShell("drawingArea", "npw-viewer", applicationShellWidgetClass, x_display, NULL, 0);
	plugin->top_widget = top_widget;

	Arg args[7];
	int n = 0;
	XtSetArg(args[n], XtNwidth, window->width); n++;
	XtSetArg(args[n], XtNheight, window->height); n++;
	XtSetValues(top_widget, args, n);

	Widget form = XtVaCreateWidget("form", compositeWidgetClass, top_widget, NULL);
	plugin->form = form;

	n = 0;
	XtSetArg(args[n], XtNwidth, window->width); n++;
	XtSetArg(args[n], XtNheight, window->height); n++;
	XtSetArg(args[n], XtNvisual, ws_info->visual); n++;
	XtSetArg(args[n], XtNcolormap, ws_info->colormap); n++;
	XtSetArg(args[n], XtNdepth, ws_info->depth); n++;
	XtSetValues(form, args, n);

	XSync(x_display, False);

#if USE_XEMBED
	plugin->old_window = top_widget->core.window;
	top_widget->core.window = GDK_WINDOW_XWINDOW(plugin->window->window);
	XtRegisterDrawable(x_display, GDK_WINDOW_XWINDOW(plugin->window->window), top_widget);
#else
	plugin->old_window = top_widget->core.window;
	top_widget->core.window = plugin->window;
	XtRegisterDrawable(x_display, plugin->window, top_widget);
#endif

	XtRealizeWidget(form);
	XtRealizeWidget(top_widget);
	XtManageChild(form);

	XSelectInput(x_display, XtWindow(top_widget), 0x0fffff);
	XSelectInput(x_display, XtWindow(form), 0x0fffff);

	XSync(x_display, False);
  }
  window->window = (void *)XtWindow(plugin->form);

  NPError ret = g_NPP_SetWindow(plugin, window);
  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_INVALID);
}

// NPP_GetValue
static NPError
g_NPP_GetValue(PluginInstance *plugin, NPPVariable variable, void *value)
{
  if (plugin == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  if (plugin_funcs.getvalue == NULL)
	return NPERR_INVALID_FUNCTABLE_ERROR;

  D(bug("NPP_GetValue instance=%p, variable=%d\n", plugin->instance, variable));
  NPError ret = plugin_funcs.getvalue(plugin->instance, variable, value);
  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

static int handle_NPP_GetValue(rpc_connection_t *connection)
{
  int error;
  uint32_t plugin_id;
  int32_t variable;

  error = rpc_method_get_args(connection,
							  RPC_TYPE_UINT32, &plugin_id,
							  RPC_TYPE_INT32, &variable,
							  RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPP_GetValue variable\n");
	return error;
  }

  NPError ret = NPERR_GENERIC_ERROR;
  int variable_type = rpc_type_of_NPPVariable(variable);

  switch (variable_type) {
  case RPC_TYPE_STRING:
	{
	  char *str = NULL;
	  ret = g_NPP_GetValue(id_lookup(plugin_id), variable, (void *)&str);
	  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_STRING, str, RPC_TYPE_INVALID);
	}
  case RPC_TYPE_INT32:
	{
	  uint32_t n = 0;
	  ret = g_NPP_GetValue(id_lookup(plugin_id), variable, (void *)&n);
	  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_INT32, n, RPC_TYPE_INVALID);
	}
  case RPC_TYPE_BOOLEAN:
	{
	  PRBool b = FALSE;
	  switch (variable) {
	  case NPPVpluginNeedsXEmbed:
#if USE_XEMBED
		b = PR_TRUE;
#else
		D(bug(" XEmbed not supported yet\n"));
		b = PR_FALSE;
#endif
		ret = NPERR_NO_ERROR;
		break;
	  default:
		ret = g_NPP_GetValue(id_lookup(plugin_id), variable, (void *)&b);
		break;
	  }
	  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_BOOLEAN, b, RPC_TYPE_INVALID);
	}
  }

  abort();
}

// NPP_URLNotify
static void
g_NPP_URLNotify(PluginInstance *plugin, const char *url, NPReason reason, void *notifyData)
{
  if (plugin == NULL)
	return;

  if (plugin_funcs.urlnotify == NULL)
	return;

  D(bug("NPP_URLNotify instance=%p, url='%s', reason=%s, notifyData=%p\n",
		plugin->instance, url, string_of_NPReason(reason), notifyData));
  plugin_funcs.urlnotify(plugin->instance, url, reason, notifyData);
  D(bug(" done\n"));
}

static int handle_NPP_URLNotify(rpc_connection_t *connection)
{
  int error;
  uint32_t plugin_id;
  char *url;
  int32_t reason;
  void *notifyData;

  error = rpc_method_get_args(connection,
							  RPC_TYPE_UINT32, &plugin_id,
							  RPC_TYPE_STRING, &url,
							  RPC_TYPE_INT32, &reason,
							  RPC_TYPE_UINT32, &notifyData,
							  RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPP_URLNotify arguments\n");
	return error;
  }

  g_NPP_URLNotify(id_lookup(plugin_id), url, reason, notifyData);

  if (url)
	free(url);

  return RPC_ERROR_NO_ERROR;
}

// NPP_NewStream
static NPError
g_NPP_NewStream(PluginInstance *plugin, NPMIMEType type, NPStream *stream, NPBool seekable, uint16 *stype)
{
  if (plugin == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  if (plugin_funcs.newstream == NULL)
	return NPERR_INVALID_FUNCTABLE_ERROR;

  D(bug("NPP_NewStream instance=%p, stream=%p, type='%s', seekable=%d, stype=%s\n",
		plugin->instance, stream, type, seekable, string_of_NPStreamType(*stype)));
  NPError ret = plugin_funcs.newstream(plugin->instance, type, stream, seekable, stype);
  D(bug(" return: %d [%s], stype=%s\n", ret, string_of_NPError(ret), string_of_NPStreamType(*stype)));
  return ret;
}

static int handle_NPP_NewStream(rpc_connection_t *connection)
{
  int error;
  uint32_t plugin_id, seekable, v_stype;
  NPMIMEType type;
  NPStream *stream;
  uint16 stype = NP_NORMAL;

  if ((stream = malloc(sizeof(*stream))) == NULL)
	return RPC_ERROR_NO_MEMORY;
  memset(stream, 0, sizeof(*stream));
  int stream_id = id_create(stream);
  stream->ndata = (void *)(uintptr_t)stream_id;

  error = rpc_method_get_args(connection,
							  RPC_TYPE_UINT32, &plugin_id,
							  RPC_TYPE_STRING, &type,
							  RPC_TYPE_STRING, &stream->url,
							  RPC_TYPE_UINT32, &stream->end,
							  RPC_TYPE_UINT32, &stream->lastmodified, // XXX not pointer safe
							  RPC_TYPE_UINT32, &stream->notifyData,
							  RPC_TYPE_BOOLEAN, &seekable,
							  RPC_TYPE_UINT32, &v_stype,
							  RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPP_NewStream arguments\n");
	return error;
  }

  if (v_stype)
	stype = v_stype;

  NPError ret = g_NPP_NewStream(id_lookup(plugin_id), type, stream, seekable, &stype);

  if (type)
	free(type);

  return rpc_method_send_reply(connection,
							   RPC_TYPE_INT32, ret,
							   RPC_TYPE_UINT32, stream_id,
							   RPC_TYPE_UINT32, stream->notifyData,
							   RPC_TYPE_UINT32, stype,
							   RPC_TYPE_INVALID);
}

// NPP_DestroyStream
static NPError
g_NPP_DestroyStream(PluginInstance *plugin, NPStream *stream, NPReason reason)
{
  if (plugin == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  if (plugin_funcs.destroystream == NULL)
	return NPERR_INVALID_FUNCTABLE_ERROR;

  if (stream == NULL)
	return NPERR_INVALID_PARAM;

  D(bug("NPP_DestroyStream instance=%p, stream=%p, reason=%s\n",
		plugin->instance, stream, string_of_NPReason(reason)));
  NPError ret = plugin_funcs.destroystream(plugin->instance, stream, reason);
  free((char *)stream->url);
  free(stream);

  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

static int handle_NPP_DestroyStream(rpc_connection_t *connection)
{
  D(bug("handle_NPP_DestroyStream\n"));

  int32_t reason;
  uint32_t plugin_id, stream_id;
  int error = rpc_method_get_args(connection,
								  RPC_TYPE_UINT32, &plugin_id,
								  RPC_TYPE_UINT32, &stream_id,
								  RPC_TYPE_INT32, &reason,
								  RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPP_DestroyStream arguments\n");
	return error;
  }

  NPError ret = g_NPP_DestroyStream(id_lookup(plugin_id), id_lookup(stream_id), reason);
  id_unlink(stream_id);

  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_INVALID);
}

// NPP_WriteReady
static int32
g_NPP_WriteReady(PluginInstance *plugin, NPStream *stream)
{
  if (plugin == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  if (plugin_funcs.writeready == NULL)
	return NPERR_INVALID_FUNCTABLE_ERROR;

  if (stream == NULL)
	return NPERR_INVALID_PARAM;

  D(bug("NPP_WriteReady instance=%p, stream=%p\n", plugin->instance, stream));
  int32 ret = plugin_funcs.writeready(plugin->instance, stream);
  D(bug(" return: %d\n", ret));
  return ret;
}

static int handle_NPP_WriteReady(rpc_connection_t *connection)
{
  D(bug("handle_NPP_WriteReady\n"));

  uint32_t plugin_id, stream_id;
  int error = rpc_method_get_args(connection,
								  RPC_TYPE_UINT32, &plugin_id,
								  RPC_TYPE_UINT32, &stream_id,
								  RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPP_WriteReady arguments\n");
	return error;
  }

  int32 ret = g_NPP_WriteReady(id_lookup(plugin_id), id_lookup(stream_id));

  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_INVALID);
}

// NPP_Write
static int32
g_NPP_Write(PluginInstance *plugin, NPStream *stream, int32 offset, int32 len, void *buf)
{
  if (plugin == NULL)
	return -1;

  if (plugin_funcs.write == NULL)
	return -1;

  if (stream == NULL)
	return -1;

  D(bug("NPP_Write instance=%p, stream=%p, offset=%d, len=%d, buf=%p\n", plugin->instance, stream, offset, len, buf));
  int32 ret = plugin_funcs.write(plugin->instance, stream, offset, len, buf);
  D(bug(" return: %d\n", ret));
  return ret;
}

static int handle_NPP_Write(rpc_connection_t *connection)
{
  unsigned char *buf;
  int32_t offset, len;
  uint32_t plugin_id, stream_id;
  int error = rpc_method_get_args(connection,
								  RPC_TYPE_UINT32, &plugin_id,
								  RPC_TYPE_UINT32, &stream_id,
								  RPC_TYPE_INT32, &offset,
								  RPC_TYPE_ARRAY, RPC_TYPE_CHAR, &len, &buf,
								  RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPP_Write arguments\n");
	return error;
  }

  int32 ret = g_NPP_Write(id_lookup(plugin_id), id_lookup(stream_id), offset, len, buf);

  if (buf)
	free(buf);

  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_INVALID);
}

// NPP_StreamAsFile
static void
g_NPP_StreamAsFile(PluginInstance *plugin, NPStream *stream, const char *fname)
{
  if (plugin == NULL)
	return;

  if (plugin_funcs.asfile == NULL)
	return;

  if (stream == NULL)
	return;

  D(bug("NPP_StreamAsFile instance=%p, stream=%p, fname='%s'\n", plugin->instance, stream, fname));
  plugin_funcs.asfile(plugin->instance, stream, fname);
  D(bug(" done\n"));
}

static int handle_NPP_StreamAsFile(rpc_connection_t *connection)
{
  char *fname;
  uint32_t plugin_id, stream_id;
  int error = rpc_method_get_args(connection,
								  RPC_TYPE_UINT32, &plugin_id,
								  RPC_TYPE_UINT32, &stream_id,
								  RPC_TYPE_STRING, &fname,
								  RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPP_StreamAsFile arguments\n");
	return error;
  }

  g_NPP_StreamAsFile(id_lookup(plugin_id), id_lookup(stream_id), fname);

  if (fname)
	free(fname);

  return RPC_ERROR_NO_ERROR;
}


/* ====================================================================== */
/* === Events processing                                              === */
/* ====================================================================== */

typedef gboolean (*GSourcePrepare)(GSource *, gint *);
typedef gboolean (*GSourceCheckFunc)(GSource *);
typedef gboolean (*GSourceDispatchFunc)(GSource *, GSourceFunc, gpointer);
typedef void (*GSourceFinalizeFunc)(GSource *);

// Xt events
static GPollFD xt_event_poll_fd;

static gboolean xt_event_prepare(GSource *source, gint *timeout)
{
  *timeout = 0;
  int mask = XtAppPending(x_app_context);
  return (mask & XtIMXEvent) != 0;
}

static gboolean xt_event_check(GSource *source)
{
  int mask = (xt_event_poll_fd.revents & G_IO_IN) ? XtAppPending(x_app_context) : 0;
  return (mask & XtIMXEvent) != 0;
}

static gboolean xt_event_dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
#if DISPATCH_ALL_XT_EVENTS
  if (XtAppPending(x_app_context))
	XtAppProcessEvent(x_app_context, XtIMAll);
#else
  if (XtAppPending(x_app_context) & (XtIMXEvent))
	XtAppProcessEvent(x_app_context, XtIMXEvent);
#endif
  return TRUE;
}

static GSourceFuncs xt_event_funcs = {
  xt_event_prepare,
  xt_event_check,
  xt_event_dispatch,
  (GSourceFinalizeFunc)g_free,
  (GSourceFunc)NULL,
  (GSourceDummyMarshal)NULL
};

// RPC events
static GPollFD rpc_event_poll_fd;

static gboolean rpc_event_prepare(GSource *source, gint *timeout)
{
  *timeout = -1;
  return FALSE;
}

static gboolean rpc_event_check(GSource *source)
{
  return rpc_wait_dispatch(g_rpc_connection, 0) > 0;
}

static gboolean rpc_event_dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
  callback(user_data);
  return TRUE;
}

static GSourceFuncs rpc_event_funcs = {
  rpc_event_prepare,
  rpc_event_check,
  rpc_event_dispatch,
  (GSourceFinalizeFunc)g_free,
  (GSourceFunc)NULL,
  (GSourceDummyMarshal)NULL
};


/* ====================================================================== */
/* === Compatibility glue for Real Player 8 plugin                    === */
/* ====================================================================== */

void *__builtin_vec_new(size_t size)
{
  return malloc(size);
}

void __builtin_vec_delete(void *ptr)
{
  if (ptr)
	free(ptr);
}

void *__builtin_new(size_t size)
{
  return malloc(size);
}

void __builtin_delete(void *ptr)
{
  if (ptr)
	free(ptr);
}

void __pure_virtual(void)
{
  npw_printf("ERROR: __pure_virtual() called\n");
  _exit(1);
}

#if defined(__i386__)
// Force use of an ELF interpreter even if this is an ELF shared object
// This is so that the RP8 glue works without trickering with another
// compat DSO in LD_PRELOAD (this binary is a special DSO itself)...
const char interp[] __attribute__((section(".interp"))) = "/lib/ld-linux.so.2";
#endif


/* ====================================================================== */
/* === Main program                                                   === */
/* ====================================================================== */

int main(int argc, char **argv)
{
  // Optimize out cases where we only want version strings & MIME info
  int print_info = 0;
  if (argc == 4 && strcmp(argv[1], "-info") == 0) {
	print_info = 1;
	--argc; ++argv;
  }

  // Xt and GTK initialization
  if (!print_info) {
	XInitThreads();
	XtToolkitInitialize();
	x_app_context = XtCreateApplicationContext();
	x_display = XtOpenDisplay(x_app_context, NULL, "npw-viewer", "npw-viewer", NULL, 0, &argc, argv);
	gtk_init(&argc, &argv);
  }

  // Sanity check args
  if (argc != 3) {
	npw_printf("ERROR: invalid number of arguments to Viewer\n");
	return 1;
  }
  const char *plugin_path = argv[1];
  D(bug("  %s\n", plugin_path));
  const char *connection_path = argv[2];
  D(bug("  Plugin connection: %s\n", connection_path));

  // Open plug-in and get exported lib functions
  void *handle = dlopen(plugin_path, RTLD_LAZY);
  if (handle == NULL) {
	npw_printf("ERROR: %s\n", dlerror());
	return 1;
  }
  {
	const char *error;
	dlerror();
	g_NP_GetMIMEDescription = (NP_GetMIMEDescriptionUPP)dlsym(handle, "NP_GetMIMEDescription");
	if ((error = dlerror()) != NULL) {
	  npw_printf("ERROR: %s\n", error);
	  return 1;
	}
	g_NP_Initialize = (NP_InitializeUPP)dlsym(handle, "NP_Initialize");
	if ((error = dlerror()) != NULL) {
	  npw_printf("ERROR: %s\n", error);
	  return 1;
	}
	g_NP_Shutdown = (NP_ShutdownUPP)dlsym(handle, "NP_Shutdown");
	if ((error = dlerror()) != NULL) {
	  npw_printf("ERROR: %s\n", error);
	  return 1;
	}
	g_NP_GetValue = (NP_GetValueUPP)dlsym(handle, "NP_GetValue");
  }

  // Print info only?
  if (print_info) {
	if (g_NP_GetValue == NULL)
	  printf("0\n\n0\n\n");
	else {
	  const char *plugin_name = NULL;
	  if (g_NP_GetValue(NULL, NPPVpluginNameString, &plugin_name) == NPERR_NO_ERROR && plugin_name)
		printf("%zd\n%s\n", strlen(plugin_name) + 1, plugin_name);
	  else
		printf("0\n\n");
	  const char *plugin_desc = NULL;
	  if (g_NP_GetValue(NULL, NPPVpluginDescriptionString, &plugin_desc) == NPERR_NO_ERROR && plugin_desc)
		printf("%zd\n%s\n", strlen(plugin_desc) + 1, plugin_desc);
	  else
		printf("0\n\n");
	}
	const char *mime_info = g_NP_GetMIMEDescription();
	if (mime_info)
	  printf("%zd\n%s\n", strlen(mime_info) + 1, mime_info);
	else
	  printf("0\n\n");
	dlclose(handle);
	return 0;
  }

  // Initialize RPC communication channel
  if (rpc_add_np_marshalers() < 0) {
	npw_printf("ERROR: failed to initialize plugin-side marshalers\n");
	return 1;
  }
  if ((g_rpc_connection = rpc_init_server(connection_path)) == NULL) {
	npw_printf("ERROR: failed to initialize plugin-side RPC server connection\n");
	return 1;
  }
  static const rpc_method_descriptor_t vtable[] = {
	{ RPC_METHOD_NP_GET_MIME_DESCRIPTION,		handle_NP_GetMIMEDescription },
	{ RPC_METHOD_NP_GET_VALUE,					handle_NP_GetValue },
	{ RPC_METHOD_NP_INITIALIZE,					handle_NP_Initialize },
	{ RPC_METHOD_NP_SHUTDOWN,					handle_NP_Shutdown },
	{ RPC_METHOD_NP_EXIT,						handle_NP_Exit },
	{ RPC_METHOD_NPP_NEW,						handle_NPP_New },
	{ RPC_METHOD_NPP_DESTROY,					handle_NPP_Destroy },
	{ RPC_METHOD_NPP_GET_VALUE,					handle_NPP_GetValue },
	{ RPC_METHOD_NPP_SET_WINDOW,				handle_NPP_SetWindow },
	{ RPC_METHOD_NPP_URL_NOTIFY,				handle_NPP_URLNotify },
	{ RPC_METHOD_NPP_NEW_STREAM,				handle_NPP_NewStream },
	{ RPC_METHOD_NPP_DESTROY_STREAM,			handle_NPP_DestroyStream },
	{ RPC_METHOD_NPP_WRITE_READY,				handle_NPP_WriteReady },
	{ RPC_METHOD_NPP_WRITE,						handle_NPP_Write },
	{ RPC_METHOD_NPP_STREAM_AS_FILE,			handle_NPP_StreamAsFile }
  };
  if (rpc_method_add_callbacks(g_rpc_connection, vtable, sizeof(vtable) / sizeof(vtable[0])) < 0) {
	npw_printf("ERROR: failed to setup NPP method callbacks\n");
	return 1;
  }

  // Initialize Xt events listener (integrate X events into GTK events loop)
  GSource *xt_source = g_source_new(&xt_event_funcs, sizeof(GSource));
  if (xt_source == NULL) {
	npw_printf("ERROR: failed to initialize Xt events listener\n");
	return 1;
  }
  g_source_set_priority(xt_source, GDK_PRIORITY_EVENTS);
  gint xt_tag = g_source_attach(xt_source, NULL);
  xt_event_poll_fd.fd = ConnectionNumber(x_display);
  xt_event_poll_fd.events = G_IO_IN;
  xt_event_poll_fd.revents = 0;
  g_main_context_add_poll(NULL, &xt_event_poll_fd, G_PRIORITY_LOW);

  // Initialize RPC events listener
  GSource *rpc_source = g_source_new(&rpc_event_funcs, sizeof(GSource));
  if (rpc_source == NULL) {
	npw_printf("ERROR: failed to initialize plugin-side RPC events listener\n");
	return 1;
  }
  g_source_set_priority(rpc_source, GDK_PRIORITY_EVENTS);
  gint rpc_tag = g_source_attach(rpc_source, NULL);
  rpc_event_poll_fd.fd = rpc_listen_socket(g_rpc_connection);
  rpc_event_poll_fd.events = G_IO_IN;
  rpc_event_poll_fd.revents = 0;
  g_source_set_callback(rpc_source, (GSourceFunc)rpc_dispatch, g_rpc_connection, NULL);
  g_main_context_add_poll(NULL, &rpc_event_poll_fd, G_PRIORITY_LOW);

  gtk_main();
  D(bug("--- EXIT ---\n"));

  g_main_context_remove_poll(NULL, &rpc_event_poll_fd);
  g_source_remove(rpc_tag);
  g_main_context_remove_poll(NULL, &xt_event_poll_fd);
  g_source_remove(xt_tag);

  if (g_user_agent)
	free(g_user_agent);
  if (g_rpc_connection)
	rpc_exit(g_rpc_connection);
  if (handle)
	dlclose(handle);
  return 0;
}
