/*
 *  npw-wrapper.c - Host Mozilla plugin (loads the actual viewer)
 *
 *  nspluginwrapper (C) 2005-2006 Gwenole Beauchesne
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "sysdeps.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/wait.h>

#include <glib.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>

#include "rpc.h"
#include "npw-rpc.h"
#include "utils.h"

#define XP_UNIX 1
#define MOZ_X11 1
#include <npapi.h>
#include <npupp.h>

#define DEBUG 1
#include "debug.h"


// XXX unimplemented functions
#define UNIMPLEMENTED() npw_printf("WARNING: Unimplemented function %s at line %d\n", __func__, __LINE__)

// Globally exported plugin ident, used by the "npconfig" tool
const NPW_PluginInfo NPW_Plugin = {
  NPW_PLUGIN_IDENT,
  NPW_DEFAULT_PLUGIN_PATH,
  0
};

// Path to plugin to use
static const char *plugin_path = NPW_Plugin.path;

// Netscape exported functions
static NPNetscapeFuncs mozilla_funcs;

// Wrapper plugin data
typedef struct {
  int initialized;
  int viewer_pid;
  int is_wrapper;
  char *name;
  char *description;
  char *formats;
} Plugin;

static Plugin g_plugin = { 0, -1, 0, NULL, NULL, NULL };

// Prototypes
static void plugin_init(int is_NP_Initialize);
static void plugin_exit(void);
static pthread_mutex_t plugin_init_lock = PTHREAD_MUTEX_INITIALIZER;


/* ====================================================================== */
/* === RPC communication                                              === */
/* ====================================================================== */

static guint g_rpc_source_id;
static GPollFD g_rpc_poll_fd;
static rpc_connection_t *g_rpc_connection = NULL;

static gboolean rpc_event_prepare(GSource *source, gint *timeout)
{
  *timeout = -1;
  return FALSE;
}

static gboolean rpc_event_check(GSource *source)
{
  return rpc_wait_dispatch(g_rpc_connection, 0) > 0;
}

static gboolean rpc_event_dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
  callback(user_data);
  return TRUE;
}


/* ====================================================================== */
/* === Browser side plug-in API                                       === */
/* ====================================================================== */

// NPN_UserAgent
static int handle_NPN_UserAgent(rpc_connection_t *connection)
{
  D(bug("handle_NPN_UserAgent\n"));

  const char *user_agent = mozilla_funcs.uagent ? mozilla_funcs.uagent(NULL) : NULL;
  return rpc_method_send_reply(connection, RPC_TYPE_STRING, user_agent, RPC_TYPE_INVALID);
}

// NPN_Status
static int handle_NPN_Status(rpc_connection_t *connection)
{
  D(bug("handle_NPN_Status\n"));

  uint32_t instance_id;
  char *message;
  int error = rpc_method_get_args(connection,
								  RPC_TYPE_UINT32, &instance_id,
								  RPC_TYPE_STRING, &message,
								  RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPN_Status arguments\n");
	return error;
  }

  NPP instance = id_lookup(instance_id);
  D(bug(" instance=%p, message='%s'\n", instance, message));

  if (mozilla_funcs.status)
	mozilla_funcs.status(instance, message);
  if (message)
	free(message);
  return RPC_ERROR_NO_ERROR;
}

// NPN_GetURL
static int handle_NPN_GetURL(rpc_connection_t *connection)
{
  D(bug("handle_NPN_GetURL\n"));

  uint32_t instance_id;
  char *url, *target;
  int error = rpc_method_get_args(connection,
								  RPC_TYPE_UINT32, &instance_id,
								  RPC_TYPE_STRING, &url,
								  RPC_TYPE_STRING, &target,
								  RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPN_GetURL arguments\n");
	return error;
  }

  NPP instance = id_lookup(instance_id);
  // XXX workaround for Real Player 8, it segfaults otherwise
  if (instance->pdata == NULL)
	instance = NULL;
  D(bug(" instance=%p, url='%s', target='%s'\n", instance, url, target));

  int ret = mozilla_funcs.geturl ? mozilla_funcs.geturl(instance, url, target) : NPERR_GENERIC_ERROR;
  if (url)
	free(url);
  if (target)
	free(target);
  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_INVALID);
}

// NPN_GetURLNotify
static int handle_NPN_GetURLNotify(rpc_connection_t *connection)
{
  D(bug("handle_NPN_GetURLNotify\n"));

  uint32_t instance_id, notify_data_id;
  char *url, *target;
  int error = rpc_method_get_args(connection,
								  RPC_TYPE_UINT32, &instance_id,
								  RPC_TYPE_STRING, &url,
								  RPC_TYPE_STRING, &target,
								  RPC_TYPE_UINT32, &notify_data_id,
								  RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: could not get NPN_GetURLNotify arguments\n");
	return error;
  }

  NPP instance = id_lookup(instance_id);
  void *notifyData = (void *)notify_data_id;
  D(bug(" instance=%p, url='%s', target='%s', notifyData=%p\n", instance, url, target, notifyData));

  int ret = NPERR_GENERIC_ERROR;
  if (mozilla_funcs.geturlnotify)
	ret = mozilla_funcs.geturlnotify(instance, url, target, notifyData);
  if (url)
	free(url);
  if (target)
	free(target);

  return rpc_method_send_reply(connection, RPC_TYPE_INT32, ret, RPC_TYPE_INVALID);
}


/* ====================================================================== */
/* === Plug-in side data                                              === */
/* ====================================================================== */

static inline uint32_t instance_id(NPP instance)
{
  return (uintptr_t)instance->pdata;
}

static inline uint32_t stream_id(NPStream *stream)
{
  return (uintptr_t)stream->pdata;
}

// Creates a new instance of a plug-in
static NPError
invoke_NPP_New(NPMIMEType mime_type, NPP instance,
			   uint16_t mode, int16_t argc, char *argn[], char *argv[],
			   NPSavedData *saved)
{
  if (saved)
	npw_printf("WARNING: NPP_New with non-null saved arguments\n");

  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPP_NEW,
								RPC_TYPE_UINT32, id_create(instance),
								RPC_TYPE_STRING, mime_type,
								RPC_TYPE_INT32, (int32_t)mode,
								RPC_TYPE_ARRAY, RPC_TYPE_STRING, (uint32_t)argc, argn,
								RPC_TYPE_ARRAY, RPC_TYPE_STRING, (uint32_t)argc, argv,
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPP_New\n");
	return NPERR_GENERIC_ERROR;
  }

  int32_t ret;
  uint32_t instance_id;
  error = rpc_method_wait_for_reply(g_rpc_connection,
									RPC_TYPE_INT32, &ret,
									RPC_TYPE_UINT32, &instance_id,
									RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NPP_New\n");
	return NPERR_GENERIC_ERROR;
  }

  instance->pdata = (void *)(uintptr_t)instance_id;
  return ret;
}

static NPError
g_NPP_New(NPMIMEType mime_type, NPP instance,
		  uint16_t mode, int16_t argc, char *argn[], char *argv[],
		  NPSavedData *saved)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  D(bug("NPP_New instance=%p\n", instance));
  NPError ret = invoke_NPP_New(mime_type, instance, mode, argc, argn, argv, saved);
  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

// Deletes a specific instance of a plug-in
static NPError
invoke_NPP_Destroy(NPP instance, NPSavedData **save)
{
  if (id_remove(instance) < 0)
	return NPERR_GENERIC_ERROR;

  // XXX handle save area (transfer raw bytes but keep size information somewhere)
  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPP_DESTROY,
								RPC_TYPE_UINT32, instance_id(instance),
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPP_Destroy");
	return NPERR_GENERIC_ERROR;
  }

  int32_t ret;
  NPSavedData *save_area = NULL;
  error = rpc_method_wait_for_reply(g_rpc_connection,
									RPC_TYPE_INT32, &ret,
									RPC_TYPE_NP_SAVED_DATA, &save_area,
									RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NPP_Destroy\n");
	return NPERR_GENERIC_ERROR;
  }

  id_remove(instance);

  if (save)
	*save = save_area;

  return ret;
}

static NPError
g_NPP_Destroy(NPP instance, NPSavedData **save)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  D(bug("NPP_Destroy instance=%p\n", instance));
  NPError ret = invoke_NPP_Destroy(instance, save);
  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

// Tells the plug-in when a window is created, moved, sized, or destroyed
static NPError
invoke_NPP_SetWindow(NPP instance, NPWindow *window)
{
  NPRect *clipRect = &window->clipRect;
  NPSetWindowCallbackStruct *ws_info = window->ws_info;

  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPP_SET_WINDOW,
								RPC_TYPE_UINT32, instance_id(instance),
								RPC_TYPE_UINT32, (Window)window->window,
								RPC_TYPE_INT32, window->x,
								RPC_TYPE_INT32, window->y,
								RPC_TYPE_UINT32, window->width,
								RPC_TYPE_UINT32, window->height,
								RPC_TYPE_INT32, window->type,
								RPC_TYPE_UINT32, clipRect->top,
								RPC_TYPE_UINT32, clipRect->left,
								RPC_TYPE_UINT32, clipRect->bottom,
								RPC_TYPE_UINT32, clipRect->right,
								RPC_TYPE_INT32, ws_info->type,
								RPC_TYPE_UINT32, ws_info->depth,
								RPC_TYPE_INT32, ws_info->visual->class,
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPP_SetWindow\n");
	return NPERR_GENERIC_ERROR;
  }

  int32_t ret;
  error = rpc_method_wait_for_reply(g_rpc_connection,
									RPC_TYPE_INT32, &ret,
									RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NP_SetWindow [%d]\n", error);
	return NPERR_GENERIC_ERROR;
  }

  return ret;
}

static NPError
g_NPP_SetWindow(NPP instance, NPWindow *window)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  if (window == NULL)
	return NPERR_NO_ERROR;

  D(bug("NPP_SetWindow instance=%p\n", instance));
  NPError ret = invoke_NPP_SetWindow(instance, window);
  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

// Allows the browser to query the plug-in for information
static NPError
invoke_NPP_GetValue(NPP instance, NPPVariable variable, void *value)
{
  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPP_GET_VALUE,
								RPC_TYPE_UINT32, instance_id(instance),
								RPC_TYPE_INT32, variable,
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPP_GetValue\n");
	return NPERR_GENERIC_ERROR;
  }

  int32_t ret;
  switch (rpc_type_of_NPPVariable(variable)) {
  case RPC_TYPE_STRING:
	{
	  char *str = NULL;
	  error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_INT32, &ret, RPC_TYPE_STRING, &str, RPC_TYPE_INVALID);
	  if (error == RPC_ERROR_NO_ERROR && ret == NPERR_NO_ERROR) {
		D(bug(" value: %s\n", str));
		*((char **)value) = str;
	  }
	  break;
	}
  case RPC_TYPE_INT32:
	{
	  int32_t n = 0;
	  error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_INT32, &ret, RPC_TYPE_INT32, &n, RPC_TYPE_INVALID);
	  if (error == RPC_ERROR_NO_ERROR && ret == NPERR_NO_ERROR) {
		D(bug(" value: %d\n", n));
		*((int *)value) = n;
	  }
	  break;
	}
  case RPC_TYPE_BOOLEAN:
	{
	  uint32_t b = 0;
	  error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_INT32, &ret, RPC_TYPE_BOOLEAN, &b, RPC_TYPE_INVALID);
	  if (error == RPC_ERROR_NO_ERROR && ret == NPERR_NO_ERROR) {
		D(bug(" value: %s [%d]\n", b ? "true" : "false", b));
		*((PRBool *)value) = b;
	  }
	  break;
	}
  }

  return ret;
}

static NPError
g_NPP_GetValue(NPP instance, NPPVariable variable, void *value)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  switch (rpc_type_of_NPPVariable(variable)) {
  case RPC_TYPE_STRING:
  case RPC_TYPE_INT32:
  case RPC_TYPE_BOOLEAN:
	break;
  default:
	npw_printf("WARNING: unhandled variable %d for NPP_GetValue\n", variable);
	return NPERR_INVALID_PARAM;
  }

  D(bug("NPP_GetValue instance=%p, variable=%d\n", instance, variable));
  NPError ret = invoke_NPP_GetValue(instance, variable, value);
  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

// Sets information about the plug-in
static NPError
invoke_NPP_SetValue(NPP instance, NPPVariable variable, void *value)
{
  UNIMPLEMENTED();

  return NPERR_GENERIC_ERROR;
}

static NPError
g_NPP_SetValue(NPP instance, NPPVariable variable, void *value)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  D(bug("NPP_SetValue instance=%p, variable=%d\n", instance, variable));
  NPError ret = invoke_NPP_SetValue(instance, variable, value);
  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return NPERR_GENERIC_ERROR;
}

// Notifies the instance of the completion of a URL request
static void
invoke_NPP_URLNotify(NPP instance, const char *url, NPReason reason, void *notifyData)
{
  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPP_URL_NOTIFY,
								RPC_TYPE_UINT32, instance_id(instance),
								RPC_TYPE_STRING, url,
								RPC_TYPE_INT32, reason,
								RPC_TYPE_UINT32, (uint32_t)notifyData,
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPP_URLNotify\n");
	return;
  }

  error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_INVALID);
  
  if (error != RPC_ERROR_NO_ERROR)
	npw_dprintf("ERROR: failed to receive ACK from NPP_URLNotify [%d]\n", error);
}

static void
g_NPP_URLNotify(NPP instance, const char *url, NPReason reason, void *notifyData)
{
  if (instance == NULL)
	return;

  D(bug("NPP_URLNotify instance=%p, url='%s', reason=%d, notifyData=%p\n", instance, url, reason, notifyData));
  invoke_NPP_URLNotify(instance, url, reason, notifyData);
  D(bug(" done\n"));
}

// Notifies a plug-in instance of a new data stream
static NPError
invoke_NPP_NewStream(NPP instance, NPMIMEType type, NPStream *stream, NPBool seekable, uint16 *stype)
{
  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPP_NEW_STREAM,
								RPC_TYPE_UINT32, instance_id(instance),
								RPC_TYPE_STRING, type,
								RPC_TYPE_STRING, stream->url,
								RPC_TYPE_UINT32, stream->end,
								RPC_TYPE_UINT32, stream->lastmodified,
								RPC_TYPE_UINT32, (uint32_t)stream->notifyData, // XXX not pointer safe
								RPC_TYPE_BOOLEAN, seekable,
								RPC_TYPE_UINT32, stype ? *stype : 0,
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPP_NewStream\n");
	return NPERR_GENERIC_ERROR;
  }

  int32_t ret;
  uint32_t stream_id, notifyData, r_stype;
  error = rpc_method_wait_for_reply(g_rpc_connection,
									RPC_TYPE_INT32, &ret,
									RPC_TYPE_UINT32, &stream_id,
									RPC_TYPE_UINT32, &notifyData,
									RPC_TYPE_UINT32, &r_stype,
									RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NPP_NewStream\n");
	return NPERR_GENERIC_ERROR;
  }

  stream->pdata = (void *)(uintptr_t)stream_id;
  stream->notifyData = (void *)notifyData;
  *stype = r_stype;
  return ret;
}

static NPError
g_NPP_NewStream(NPP instance, NPMIMEType type, NPStream *stream, NPBool seekable, uint16 *stype)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  D(bug("NPP_NewStream instance=%p\n", instance));
  NPError ret = invoke_NPP_NewStream(instance, type, stream, seekable, stype);
  D(bug(" return: %d [%s], stype=%s\n", ret, string_of_NPError(ret), string_of_NPStreamType(*stype)));
  return ret;
}

// Tells the plug-in that a stream is about to be closed or destroyed
static NPError
invoke_NPP_DestroyStream(NPP instance, NPStream *stream, NPReason reason)
{
  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPP_DESTROY_STREAM,
								RPC_TYPE_UINT32, instance_id(instance),
								RPC_TYPE_UINT32, stream_id(stream),
								RPC_TYPE_INT32, reason,
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPP_DestroyStream\n");
	return NPERR_GENERIC_ERROR;
  }

  int32_t ret;
  error = rpc_method_wait_for_reply(g_rpc_connection,
									RPC_TYPE_INT32, &ret,
									RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NPP_DestroyStream\n");
	return NPERR_GENERIC_ERROR;
  }

  return ret;
}

static NPError
g_NPP_DestroyStream(NPP instance, NPStream *stream, NPReason reason)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  D(bug("NPP_DestroyStream instance=%p\n", instance));
  NPError ret = invoke_NPP_DestroyStream(instance, stream, reason);
  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

// Provides a local file name for the data from a stream
static void
invoke_NPP_StreamAsFile(NPP instance, NPStream *stream, const char *fname)
{
  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPP_STREAM_AS_FILE,
								RPC_TYPE_UINT32, instance_id(instance),
								RPC_TYPE_UINT32, stream_id(stream),
								RPC_TYPE_STRING, fname,
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPP_StreamAsFile\n");
	return;
  }

  error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR)
	npw_printf("ERROR: failed to receive ACK from NPP_StreamAsFile\n");
}

static void
g_NPP_StreamAsFile(NPP instance, NPStream *stream, const char *fname)
{
  if (instance == NULL)
	return;

  D(bug("NPP_StreamAsFile instance=%p\n", instance));
  invoke_NPP_StreamAsFile(instance, stream, fname);
  D(bug(" done\n"));
}

// Determines maximum number of bytes that the plug-in can consume
static int32
invoke_NPP_WriteReady(NPP instance, NPStream *stream)
{
  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPP_WRITE_READY,
								RPC_TYPE_UINT32, instance_id(instance),
								RPC_TYPE_UINT32, stream_id(stream),
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPP_WriteReady\n");
	return 0;
  }

  int32_t ret;
  error = rpc_method_wait_for_reply(g_rpc_connection,
									RPC_TYPE_INT32, &ret,
									RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NPP_WriteReady\n");
	return 0;
  }

  return ret;
}

static int32
g_NPP_WriteReady(NPP instance, NPStream *stream)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  D(bug("NPP_WriteReady instance=%p\n", instance));
  int32 ret = invoke_NPP_WriteReady(instance, stream);
  D(bug(" return: %d\n", ret));
  return ret;
}


// Delivers data to a plug-in instance
static int32
invoke_NPP_Write(NPP instance, NPStream *stream, int32 offset, int32 len, void *buf)
{
  int error = rpc_method_invoke(g_rpc_connection,
								RPC_METHOD_NPP_WRITE,
								RPC_TYPE_UINT32, instance_id(instance),
								RPC_TYPE_UINT32, stream_id(stream),
								RPC_TYPE_INT32, offset,
								RPC_TYPE_ARRAY, RPC_TYPE_CHAR, len, buf,
								RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NPP_Write\n");
	return -1;
  }

  int32_t ret;
  error = rpc_method_wait_for_reply(g_rpc_connection,
									RPC_TYPE_INT32, &ret,
									RPC_TYPE_INVALID);

  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NPP_Write\n");
	return -1;
  }

  return ret;
}

static int32
g_NPP_Write(NPP instance, NPStream *stream, int32 offset, int32 len, void *buf)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  D(bug("NPP_Write instance=%p\n", instance));
  int32 ret = invoke_NPP_Write(instance, stream, offset, len, buf);
  D(bug(" return: %d\n", ret));
  return ret;
}


// Requests a platform-specific print operation for an embedded or full-screen plug-in
static void invoke_NPP_Print(NPP instance, NPPrint *PrintInfo)
{
  UNIMPLEMENTED();
}

static void g_NPP_Print(NPP instance, NPPrint *PrintInfo)
{
  if (instance == NULL)
	return;

  D(bug("NPP_Print instance=%p\n", instance));
  invoke_NPP_Print(instance, PrintInfo);
  D(bug(" done\n"));
}

// Delivers a platform-specific window event to the instance
static int16 invoke_NPP_HandleEvent(NPP instance, void *event)
{
  UNIMPLEMENTED();

  return 0;
}

static int16 g_NPP_HandleEvent(NPP instance, void *event)
{
  if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

  D(bug("NPP_HandleEvent instance=%p\n", instance));
  int16 ret = invoke_NPP_HandleEvent(instance, event);
  D(bug(" return: ret\n", ret));
  return ret;
}

// Allows the browser to query the plug-in for information
NPError
NP_GetValue(void *future, NPPVariable variable, void *value)
{
  D(bug("NP_GetValue\n"));

  if (g_plugin.initialized == 0)
	plugin_init(0);
  if (g_plugin.initialized <= 0)
	return NPERR_GENERIC_ERROR;

  char *str = NULL;
  int ret = NPERR_GENERIC_ERROR;
  switch (variable) {
  case NPPVpluginNameString:
	if (g_plugin.is_wrapper) {
	  str = "Netscape 4 Plugins Wrapper";
	  ret = NPERR_NO_ERROR;
	}
	else if (g_plugin.name) {
	  str = g_plugin.name;
	  ret = NPERR_NO_ERROR;
	}
	break;
  case NPPVpluginDescriptionString:
	if (g_plugin.is_wrapper) {
	  str =
		"<a href=\"http://www.gibix.net/projects/nspluginwrapper/\">nspluginwrapper</a> "
		NPW_VERSION
		" adds support for " TARGET_ARCH " compiled Netscape 4 plugins.<br>"
		"This is <b>beta</b> software available under the terms of the GNU General Public License.<br>"
		;
	  ret = NPERR_NO_ERROR;
	}
	else if (g_plugin.description) {
	  str = g_plugin.description;
	  ret = NPERR_NO_ERROR;
	}
	break;
  default:
	return NPERR_INVALID_PARAM;
  }
  *((char **)value) = str;

  D(bug(" return: %d ['%s']\n", ret, str));
  return ret;
}

// Allows the browser to query the plug-in supported formats
char *
NP_GetMIMEDescription(void)
{
  D(bug("NP_GetMIMEDescription\n"));

  if (g_plugin.initialized == 0)
	plugin_init(0);
  if (g_plugin.initialized <= 0)
	return NULL;

  if (g_plugin.is_wrapper)
	return "unknown/mime-type:none:Do not open";

  D(bug(" formats: '%s'\n", g_plugin.formats));
  return g_plugin.formats;
}

// Provides global initialization for a plug-in
NPError
NP_Initialize(NPNetscapeFuncs *moz_funcs, NPPluginFuncs *plugin_funcs)
{
  D(bug("NP_Initialize\n"));

  if (moz_funcs == NULL || plugin_funcs == NULL)
	return NPERR_INVALID_FUNCTABLE_ERROR;

  if ((moz_funcs->version >> 8) > NP_VERSION_MAJOR)
    return NPERR_INCOMPATIBLE_VERSION_ERROR;
  if (moz_funcs->size < sizeof(NPNetscapeFuncs))
    return NPERR_INVALID_FUNCTABLE_ERROR;
  if (plugin_funcs->size < sizeof(NPPluginFuncs))
    return NPERR_INVALID_FUNCTABLE_ERROR;
  if (g_plugin.is_wrapper)
	return NPERR_NO_ERROR;

  if (g_plugin.initialized == 0 || g_plugin.initialized == 1)
	plugin_init(1);
  if (g_plugin.initialized <= 0)
	return NPERR_GENERIC_ERROR;

  memcpy(&mozilla_funcs, moz_funcs, sizeof(*moz_funcs));

  NPNToolkitType toolkit = 0;
  int error = mozilla_funcs.getvalue(NULL, NPNVToolkit, (void *)&toolkit);
  if (error != NPERR_NO_ERROR || toolkit != NPNVGtk2) {
	npw_printf("ERROR: your browser does not support GTK+ 2 peer\n");
	return NPERR_INCOMPATIBLE_VERSION_ERROR;
  }

  memset(plugin_funcs, 0, sizeof(*plugin_funcs));
  plugin_funcs->size = sizeof(NPPluginFuncs);
  plugin_funcs->version = (NP_VERSION_MAJOR << 8) + NP_VERSION_MINOR;
  plugin_funcs->newp = NewNPP_NewProc(g_NPP_New);
  plugin_funcs->destroy = NewNPP_DestroyProc(g_NPP_Destroy);
  plugin_funcs->setwindow = NewNPP_SetWindowProc(g_NPP_SetWindow);
  plugin_funcs->newstream = NewNPP_NewStreamProc(g_NPP_NewStream);
  plugin_funcs->destroystream = NewNPP_DestroyStreamProc(g_NPP_DestroyStream);
  plugin_funcs->asfile = NewNPP_StreamAsFileProc(g_NPP_StreamAsFile);
  plugin_funcs->writeready = NewNPP_WriteReadyProc(g_NPP_WriteReady);
  plugin_funcs->write = NewNPP_WriteProc(g_NPP_Write);
  plugin_funcs->print = NewNPP_PrintProc(g_NPP_Print);
  plugin_funcs->event = NewNPP_HandleEventProc(g_NPP_HandleEvent);
  plugin_funcs->urlnotify = NewNPP_URLNotifyProc(g_NPP_URLNotify);
  plugin_funcs->javaClass = NULL;
  plugin_funcs->getvalue = NewNPP_GetValueProc(g_NPP_GetValue);
  plugin_funcs->setvalue = NewNPP_SetValueProc(g_NPP_SetValue);

  error = rpc_method_invoke(g_rpc_connection, RPC_METHOD_NP_INITIALIZE, RPC_TYPE_INVALID);
  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NP_Initialize\n");
	return NPERR_GENERIC_ERROR;
  }

  int32_t ret;
  error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_INT32, &ret, RPC_TYPE_INVALID);
  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NP_Initialize\n");
	return NPERR_GENERIC_ERROR;
  }

  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}

// Provides global deinitialization for a plug-in
NPError
NP_Shutdown(void)
{
  D(bug("NP_Shutdown\n"));

  if (g_plugin.is_wrapper)
	return NPERR_NO_ERROR;

  int error = rpc_method_invoke(g_rpc_connection, RPC_METHOD_NP_SHUTDOWN, RPC_TYPE_INVALID);
  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to call NP_Shutdown\n");
	return NPERR_GENERIC_ERROR;
  }

  int32_t ret;
  error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_INT32, &ret, RPC_TYPE_INVALID);
  if (error != RPC_ERROR_NO_ERROR) {
	npw_printf("ERROR: failed to receive reply from NP_Shutdown\n");
	return NPERR_GENERIC_ERROR;
  }

  plugin_exit();

  D(bug(" return: %d [%s]\n", ret, string_of_NPError(ret)));
  return ret;
}


/* ====================================================================== */
/* === Plug initialization                                            === */
/* ====================================================================== */

// Initialize wrapper plugin and execute viewer
static void do_plugin_init(int is_NP_Initialize)
{
  if (g_plugin.initialized < 0)
	return;
  g_plugin.initialized = -1;

  D(bug("plugin_init for %s\n", plugin_path));
  if (strcmp(plugin_path, NPW_DEFAULT_PLUGIN_PATH) == 0) {
	g_plugin.is_wrapper = 1;
	g_plugin.initialized = 1 + is_NP_Initialize;
	return;
  }

  static const char *plugin_file_name = NULL;
  if (plugin_file_name == NULL) {
	const char *p;
	for (p = &plugin_path[strlen(plugin_path) - 1]; p > plugin_path; p--) {
	  if (*p == '/') {
		plugin_file_name = p + 1;
		break;
	  }
	}
	if (plugin_file_name == NULL)
	  return;
  }

  static int init_count = 0;
  ++init_count;
  char connection_path[128];
  sprintf(connection_path, "%s/%s/%d-%d", NPW_CONNECTION_PATH, plugin_file_name, getpid(), init_count);

  // Cache MIME info and plugin name/description
  if (!is_NP_Initialize && g_plugin.name == NULL && g_plugin.description == NULL && g_plugin.formats == NULL) {
	char command[1024];
	if (snprintf(command, sizeof(command), "%s/%s -info %s 0", NPW_LIBDIR, NPW_VIEWER, plugin_path) >= sizeof(command))
	  return;
	FILE *viewer_fp = popen(command, "r");
	if (viewer_fp == NULL)
	  return;
	char **strings[] = { &g_plugin.name, &g_plugin.description, &g_plugin.formats, NULL };
	int i, error = 0;
	for (i = 0; strings[i] != NULL; i++) {
	  int len;
	  if (fscanf(viewer_fp, "%d\n", &len) != 1) {
		error = 1;
		break;
	  }
	  char *str = malloc(len + 1);
	  if (len == 0)
		str[0] = '\0';
	  else {
		if (fgets(str, len, viewer_fp) == NULL) {
		  error = 1;
		  break;
		}
	  }
	  *(strings[i]) = str;
	}
	pclose(viewer_fp);
	if (error == 0)
	  g_plugin.initialized = 1;
	return;
  }

  // Start plug-in viewer
  if ((g_plugin.viewer_pid = fork()) == 0) {
	char *argv[8];
	int argc = 0;

	argv[argc++] = NPW_VIEWER;
	argv[argc++] = (char *)plugin_path;
	argv[argc++] = connection_path;
	argv[argc] = NULL;

	execv(NPW_LIBDIR "/" NPW_VIEWER, argv);
	npw_printf("ERROR: failed to execute NSPlugin viewer\n");
	return;
  }

  // Initialize browser-side RPC communication channel
  if (rpc_add_np_marshalers() < 0) {
	npw_printf("ERROR: failed to initialize browser-side marshalers\n");
	return;
  }
  if ((g_rpc_connection = rpc_init_client(connection_path)) == NULL) {
	npw_printf("ERROR: failed to initialize plugin-side RPC client connection\n");
	return;
  }
  static const rpc_method_descriptor_t vtable[] = {
	{ RPC_METHOD_NPN_USER_AGENT,						handle_NPN_UserAgent },
	{ RPC_METHOD_NPN_GET_URL,							handle_NPN_GetURL },
	{ RPC_METHOD_NPN_GET_URL_NOTIFY,					handle_NPN_GetURLNotify },
	{ RPC_METHOD_NPN_STATUS,							handle_NPN_Status }
  };
  if (rpc_method_add_callbacks(g_rpc_connection, vtable, sizeof(vtable) / sizeof(vtable[0])) < 0) {
	npw_printf("ERROR: failed to setup NPN method callbacks\n");
	return;
  }

  // Initialize RPC events listener
  static GSourceFuncs rpc_event_funcs = {
	rpc_event_prepare,
	rpc_event_check,
	rpc_event_dispatch,
	(void (*)(GSource *))g_free,
	(GSourceFunc)NULL,
	(GSourceDummyMarshal)NULL
  };
  GSource *rpc_source = g_source_new(&rpc_event_funcs, sizeof(GSource));
  if (rpc_source == NULL) {
	npw_printf("ERROR: failed to initialize brower-side RPC events listener\n");
	return;
  }
  g_source_set_priority(rpc_source, G_PRIORITY_DEFAULT);
  g_rpc_source_id = g_source_attach(rpc_source, NULL);
  g_rpc_poll_fd.fd = rpc_socket(g_rpc_connection);
  g_rpc_poll_fd.events = G_IO_IN;
  g_rpc_poll_fd.revents = 0;
  g_source_set_callback(rpc_source, (GSourceFunc)rpc_dispatch, g_rpc_connection, NULL);
  g_main_context_add_poll(NULL, &g_rpc_poll_fd, G_PRIORITY_LOW);

  g_plugin.initialized = 1 + is_NP_Initialize;
  D(bug("--- INIT ---\n"));
}

static void plugin_init(int is_NP_Initialize)
{
  pthread_mutex_lock(&plugin_init_lock);
  do_plugin_init(is_NP_Initialize);
  pthread_mutex_unlock(&plugin_init_lock);
}

// Kill NSPlugin Viewer process
static void do_plugin_exit(void)
{
  D(bug("plugin_exit\n"));

  if (g_rpc_source_id) {
	g_main_context_remove_poll(NULL, &g_rpc_poll_fd);
	g_source_remove(g_rpc_source_id);
	g_rpc_source_id = 0;
  }

  if (g_rpc_connection) {
	if (rpc_method_invoke(g_rpc_connection, RPC_METHOD_NP_EXIT, RPC_TYPE_INVALID) != RPC_ERROR_NO_ERROR)
	  npw_printf("ERROR: failed to call NP_Exit\n");
	rpc_exit(g_rpc_connection);
	g_rpc_connection = NULL;
  }

  if (g_plugin.viewer_pid != -1) {
	int status;
	waitpid(g_plugin.viewer_pid, &status, 0);
	g_plugin.viewer_pid = -1;
  }

  g_plugin.initialized = 0;
}

static void plugin_exit(void)
{
  pthread_mutex_lock(&plugin_init_lock);
  do_plugin_exit();
  pthread_mutex_unlock(&plugin_init_lock);
}

static void __attribute__((destructor)) plugin_exit_sentinel(void)
{
  plugin_exit();

  if (g_plugin.formats) {
	free(g_plugin.formats);
	g_plugin.formats = NULL;
  }

  if (g_plugin.name) {
	free(g_plugin.name);
	g_plugin.name = NULL;
  }

  if (g_plugin.description) {
	free(g_plugin.description);
	g_plugin.description = NULL;
  }
}
