/*
 *  utils.c - Utility functions
 *
 *  nspluginwrapper (C) 2005-2006 Gwenole Beauchesne
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "sysdeps.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "utils.h"

#ifndef TEST_UTILS
#define XP_UNIX 1
#define MOZ_X11 1
#include <npapi.h>
#endif

#define DEBUG 1
#include "debug.h"


/* ====================================================================== */
/* === Hashes                                                         === */
/* ====================================================================== */

#define N_CELLS_PER_SLOT 8

typedef struct _Slot {
  int id_base;
  void *values[N_CELLS_PER_SLOT];
  struct _Slot *next;
} Slot;

static int g_current_id = 0;
static Slot *g_ids = NULL;

static Slot *id_lookup_slot(int id, Slot **pslp)
{
  Slot *psl = NULL;
  Slot *sl = g_ids;

  const int n_slots = id / N_CELLS_PER_SLOT;

  if (sl) {
	int i;
	for (i = 0; sl && sl->id_base != n_slots && i < n_slots; i++) {
	  psl = sl;
	  sl = sl->next;
	}
  }

  if (sl && sl->id_base != n_slots) {
	D(bug("ERROR: failed to find appropriate slot for id %d\n", id));
	sl = NULL;
  }

  if (pslp)
	*pslp = psl;

  return sl;
}

int id_link(void *ptr, int id)
{
  Slot *psl;
  Slot *sl = id_lookup_slot(id, &psl);

  if (sl == NULL) {
	sl = malloc(sizeof(*sl));
	memset(sl->values, 0, sizeof(sl->values));
	sl->id_base = id / N_CELLS_PER_SLOT;
	sl->next = NULL;
	if (psl)
	  psl->next = sl;
	else
	  g_ids = sl;
  }

  sl->values[id % N_CELLS_PER_SLOT] = ptr;
  return id;
}

int id_unlink(int id)
{
  Slot *psl, *sl = id_lookup_slot(id, &psl);
  if (sl == NULL)
	return -1;

  sl->values[id % N_CELLS_PER_SLOT] = NULL;

  int i;
  for (i = 0; i < N_CELLS_PER_SLOT; i++)
	if (sl->values[i])
	  return 0;

  if (sl == g_ids)
	g_ids = sl->next;

  if (psl)
	psl->next = sl->next;

  free(sl);
  return 0;
}

int id_create(void *ptr)
{
  ++g_current_id;
  return id_link(ptr, g_current_id);
}

int id_remove(void *ptr)
{
  Slot *sl = g_ids;

  if (sl == NULL)
	return -1;

  int n_slots = 0;

  while (sl) {
	int i;
	for (i = 0; i < N_CELLS_PER_SLOT; i++) {
	  if (sl->values[i] == ptr)
		return id_unlink((n_slots * N_CELLS_PER_SLOT) + i);
	}
	n_slots++;
	sl = sl->next;
  }

  return -2;
}

void *id_lookup(int id)
{
  Slot *sl = id_lookup_slot(id, NULL);
  if (sl == NULL)
	return NULL;

  return sl->values[id % N_CELLS_PER_SLOT];
}


/* ====================================================================== */
/* === String expansions                                              === */
/* ====================================================================== */

#ifndef TEST_UTILS
const char *string_of_NPError(int error)
{
  const char *str;

  switch ((NPError)error) {
#define _(VAL) case VAL: str = #VAL; break;
	_(NPERR_NO_ERROR);
	_(NPERR_GENERIC_ERROR);
	_(NPERR_INVALID_INSTANCE_ERROR);
	_(NPERR_INVALID_FUNCTABLE_ERROR);
	_(NPERR_MODULE_LOAD_FAILED_ERROR);
	_(NPERR_OUT_OF_MEMORY_ERROR);
	_(NPERR_INVALID_PLUGIN_ERROR);
	_(NPERR_INVALID_PLUGIN_DIR_ERROR);
	_(NPERR_INCOMPATIBLE_VERSION_ERROR);
	_(NPERR_INVALID_PARAM);
	_(NPERR_INVALID_URL);
	_(NPERR_FILE_NOT_FOUND);
	_(NPERR_NO_DATA);
	_(NPERR_STREAM_NOT_SEEKABLE);
#undef _
  default:
	str = "<unknown error";
	break;
  }

  return str;
}

const char *string_of_NPReason(int reason)
{
  const char *str;

  switch ((NPReason)reason) {
#define _(VAL) case VAL: str = #VAL; break;
    _(NPRES_DONE);
    _(NPRES_NETWORK_ERR);
    _(NPRES_USER_BREAK);
#undef _
  default:
	str = "<unknown reason>";
	break;
  }

  return str;
}

const char *string_of_NPStreamType(int stype)
{
  const char *str;

  switch (stype) {
#define _(VAL) case VAL: str = #VAL; break;
    _(NP_NORMAL);
    _(NP_SEEK);
    _(NP_ASFILE);
    _(NP_ASFILEONLY);
#undef _
  default:
	str = "<unknown stream type>";
	break;
  }

  return str;
}
#endif


/* ====================================================================== */
/* === Misc utility functions                                         === */
/* ====================================================================== */

void npw_printf(const char *format, ...)
{
  va_list args;
  va_start(args, format);
  fprintf(stderr, "*** NSPlugin %s *** ", NPW_COMPONENT_NAME);
  vfprintf(stderr, format, args);
  va_end(args);
}


/* ====================================================================== */
/* === Test Program                                                   === */
/* ====================================================================== */

#ifdef TEST_UTILS
int main(void)
{
  char *str;
  int i, id;

#define N_STRINGS ((2 * N_CELLS_PER_SLOT) + 3)
  char *strings[N_STRINGS];
  int ids[N_STRINGS];

  for (i = 0; i < N_STRINGS; i++) {
	str = malloc(10);
	sprintf(str, "%d", i);
	strings[i] = str;
	id = id_create(str);
	if (id < 0) {
	  fprintf(stderr, "ERROR: failed to allocate ID for '%s'\n", str);
	  return 1;
	}
	ids[i] = id;
  }

  // basic lookup
  id = ids[N_CELLS_PER_SLOT / 2];
  str = id_lookup(id);
  printf("str(%d) : '%s'\n", id, str);

  // basic unlink
  id = ids[N_CELLS_PER_SLOT];
  if (id_unlink(id) < 0) {
	fprintf(stderr, "ERROR: failed to unlink ID %d\n", id);
	return 1;
  }
  ids[N_CELLS_PER_SLOT] = 0;

  // remove slot 1
  for (i = 0; i < N_CELLS_PER_SLOT; i++) {
	id = ids[N_CELLS_PER_SLOT + i];
	if (id && id_unlink(id) < 0) {
	  fprintf(stderr, "ERROR: failed to unlink ID %d from slot 1\n", id);
	  return 1;
	}
	ids[N_CELLS_PER_SLOT + i] = 0;
  }

  // basic lookup after slot removal
  id = ids[2 * N_CELLS_PER_SLOT];
  str = id_lookup(id);
  printf("str(%d) : '%s'\n", id, str);

  // check slot 1 was removed and slots 0 & 2 linked together
  for (i = 0; i < N_STRINGS; i++) {
	id = ids[i];
	if (id && id_unlink(id) < 0) {
	  fprintf(stderr, "ERROR: failed to unlink ID %d for final cleanup\n", id);
	  return 1;
	}
  }

  if (g_ids != NULL) {
	fprintf(stderr, "ERROR: memory leak\n");
	return 1;
  }

  for (i = 0; i < N_STRINGS; i++)
	free(strings[i]);

  return 0;
}
#endif
